
;
; Sooper Dooper (I hope) memory manager.
; Rick Grehan
; BYTE Magazine
;
; Caveat:
; I've tried to test this as best as I can, but things have been
; crazy around here so I can't claim that I got all the problems.
; Drop me a note if you have troubles and I'll be glad to put some
; time into tracking them down.  I'm fairly confident of the
; greater part of this (i.e., the part that doesn't swap) since
; that's based on some earlier work I did.  The most troubles
; will likely show up in the EMS and disk swapping routines.
;
; The interface routines are built with the assumption that
; you're using Tubo C, small model.  That's right, small model.
; You can easily build far pointers and get all the memory you
; need that way.  If you want to change to other models, you'll
; have to alter the stack offsets in the interface words.  Shouldn't
; be too hard.
; Hey, it's no VROOOM, but the price is right.
;
; Thanks
; Rick Grehan
;
; P.S. Sorry it took so long to get this one out.  This stuff
; isn't getting any easier.
;

	.ALPHA
	.MODEL SMALL
	.DATA
;CONSTANTS
PREVPTR		EQU	6	;M-node layout
NEXTPTR		EQU	4
BLKLENGTH	EQU	2
BLKBASE		EQU	0

;INITIALIZED DATA STORAGE
EMSDEVNAM	DB	"EMMXXXX0"
EMSDEVNAMLEN	EQU	8

;UNINITIALIZED DATA STORAGE BEGINS
	.DATA?
SWAPNAME	DB	40 DUP (?) ;Space for swapfile name
MYEMSHANDLE	DW	?	;Handle into EMS
EMSTHERE	DB	?	;Flag indicating presence of EMS
EMSBLKSFREE	DW	?	;# of free EMS 128-byte blocks
EMSPFSEG	DW	?	;Segment where EMS phys pages map
SWAPHANDLE	DW	?	;Handle for swap file
NEXTBPTR	DW	?	;Next block pointer - used in swapping
DISKAVL		DW	?	;Head of available list for disk
SWAPDEST	DB	?	;Swap destination:
				; 1 = ems
				; 2 = disk
DRTYFLAG	DB	?	;Used during compaction
NUMNODELST	DW	?	;Head of in-use mnode list
NUMNODETOP	DW	?	;Top of in-use m-node list
FREEMNODELST	DW	?	;Start of free mnode list
CURRMAT		DW	?	;Current MAT entry
EMSAVL		DW	?	;Avail list MAT
FREE_START	DW	?	;Starting seg. of free memory
HEAP_START	DW	?	;Starting seg. of heap area
HEAP_LEFT	DW	? 	;# of paragraphs in heap.
;
;	>>>>TURBO C INTERFACE ROUTINES FOLLOW
;       >>>>WRITTEN USING SMALL MODEL
;
	.CODE

	PUBLIC	_initmem
;_initmem
; This routine initializes the memory management package.
; Call it with the number of mnodes you want to
; allocate.  And a pointer to the swap file name string.
; The routine returns results in mem_struct
; (which should be passed in as a pointer).
; Format of mem_struct:
;   {	unsigned int convparas;	 # of paragraphs of conventional memory
;	unsigned long emsparas;	 # of paragraphs of EMS memory
;       int fflag;		 0 = swap file open ok
;   }
;
; Call looks like:
;   initmem(pblocks,"SWAPFILE.DAT",&mem_struct)
;
NUMNODES =	[BP+4]
SFPTR	=	[BP+6]
MSPTR	=	[BP+8]
MSPCONVPARAS	=	[SI]
MSPEMSPARAS1	=	[SI+2]
MSPEMSPARAS2	=	[SI+4]
MSPDISKOK	=	[SI+6]
_initmem:
	PUSH	BP
	MOV	BP,SP
	PUSH	SI
	PUSH	DI
;Set up memory struct pointer
	MOV	SI,MSPTR
	XOR	AX,AX
	MOV	WORD PTR MSPCONVPARAS,AX
	MOV	WORD PTR MSPEMSPARAS1,AX
	MOV	WORD PTR MSPEMSPARAS2,AX
	DEC	AX
	MOV	WORD PTR MSPDISKOK,AX
;Find out how much memory is available in conventional memory.
;Ask for more than you know you deserve.
	MOV	AH,48H
	MOV     BX,0FFFFH
	INT	21H
;We will assume failure.  BX will contain # of paragraphs actually
;available.  So, we call the interrupt again to claim the largest
;we can.  ***NOTE: You may want to modify this portionof code to
;alter the amount of conventional memory consumed by the manager.
	MOV	AH,48H
	INT	21H		;Request again
	JNC	_INIT1
_INITX:
	POP	DI		;Exit with failure
	POP	SI
	POP	BP
	RET
;Initialize the mnodes.
_INIT1:
	XCHG	AX,BX
	MOV	CX,NUMNODES
	CALL	INIT_HEAP	;Initialize heap area
	MOV	AX,HEAP_LEFT	;Return # paras of conv
	MOV     MSPCONVPARAS,AX
;Try to initialize EMS
	CALL	ISEMSHERE	;Is it there?
_INIT2:
	MOV	EMSTHERE,AL
	OR	AX,AX		;Anything there?
	JZ	_INIT5
;EMS is around.  Allocate some pages. Amount to allocate is
;smaller of amount in AX and 512
	CMP	AX,512
	JL	_INIT3
	MOV	AX,512
_INIT3:
	MOV	BX,AX		;Save # pages
	CALL	INIT_EMS
	OR	AX,AX
	JZ	_INIT4		;Jump if all ok
	XOR	AX,AX
	JMP	_INIT2
;Calculate # of free EMS paragraphs and return in structure
_INIT4:
	MOV	AX,1008		;Paragraphs per EMS page
	MUL	BX
	MOV	MSPEMSPARAS1,AX
	MOV	MSPEMSPARAS2,DX
;Copy the swap file's name into our holding area, then
;open the swap file
_INIT5:
	MOV	DX,SFPTR
	CALL	MOVESWAPNAME
	CALL	OPENSWAP
	MOV	MSPDISKOK,AX
;Return
	JMP	_INITX

	PUBLIC	_closemem
;_closemem
; Close the memory manager.  This routine:
;	1) Frees alloc'ed DOS memory.
;	2) Releases the EMS handle (if any)
;	3) Erases the swap file.
; You should call this before your program exits.
_closemem:
	PUSH	SI		;Make Turbo C happy
	PUSH	DI
	PUSH	ES
;Free the DOS memory first.
	MOV	AX,FREE_START
	MOV	ES,AX
	MOV	AH,49H		;Free allocated segment
	INT	21H
;Now free the EMS handle (if any )
	MOV	AL,EMSTHERE
	OR	AL,AL
	JZ	CLOSEMEM1
	MOV	DX,MYEMSHANDLE
	MOV	AH,45H		;Free EMS handle
	INT	67H
;Finally, close and delete the swapfile
CLOSEMEM1:
	MOV	BX,SWAPHANDLE
	MOV	AH,3EH
	INT	21H
;
	MOV	DX,OFFSET SWAPNAME
	MOV	AH,41H
	INT	21H		;Don't bother with error.
;All done, go home
	POP	ES
	POP	DI
	POP	SI
	RET
	
	PUBLIC	_gethandle
;_gethandle
; Call this function with the number of paragraphs you want to
; allocate.  It returns the handle to the mnode.
; Zero return means the request failed.
; Call with:
;  unsigned int gethandle(nparas)
;  unsigned int nparas
_gethandle:
	PUSH	BP
	MOV	BP,SP
	PUSH	SI
	PUSH	DI
	PUSH	ES
	MOV	AX,FREE_START
	MOV	ES,AX		;Set to start of heap area
	MOV	DX,[BP+4]	;Get # of paragraphs to alloc.
	CALL	ALLOC_BLK	;Do the allocation
	MOV	AX,DI
	POP	ES
	POP	DI
	POP	SI
	POP	BP
	RET

	PUBLIC	_freehandle
;_freehandle
; Call this function with the handle to an mnode you want to free.
; Routine returns with 0 if the request failed (handle was locked).
; Else returns nonzero.
; Call with:
;	unsigned int freehandle(hnum)
;	unsigned int hnum
_freehandle:
	PUSH	BP
	MOV	BP,SP
	PUSH	SI
	PUSH	DI
	MOV	AX,FREE_START
	MOV	ES,AX		;Set up ES register
	MOV	DI,[BP+4]
	CALL	FREE_BLK
	MOV	AX,DI		;Get return value
	POP	DI
	POP	SI
	POP	BP
	RET

	PUBLIC	_lockhandle
;_lockhandle
;Call this function with mnode handle you want to lock.
;Also use this to resolve a handle to a segment address.
;Routine has no effect on a handle already locked.
;Returns segment address if ok, else zero.  Zero return indicates
;block is not in conventional memory and there's no room
;to swap him in (i.e., you've overloaded your locks)
;Call with:
;	unsigned int lockhandle(hnum)
;	unisgned int hnum
_lockhandle:
	PUSH	BP
	MOV	BP,SP
	PUSH	SI
	PUSH    DI
	PUSH	ES
	MOV	AX,FREE_START
	MOV	ES,AX
	MOV	DI,[BP+4]
;Resolve MNODE number to byte address
	DEC	DI
	SHL	DI,1
	SHL	DI,1
	SHL	DI,1
;Verify that he's in memory.
	MOV	AX,ES:[DI+NEXTPTR]
	AND	AH,0C0H
	JNZ	LOCK1
;He's in memory - just lock him
LOCK0:
	OR	ES:[DI+PREVPTR],0C000H	;Set hi two bits
	MOV	AX,ES:[DI+BLKBASE]	;Get segment base
LOCKX:
	POP	ES
	POP	DI
	POP	SI
	POP	BP
	RET
;Not in memory -- we'll have to swap him in.
LOCK1:	PUSH	DI			;Save our mnode addr
	MOV	DX,ES:[DI+BLKLENGTH]
	CALL	ALLOC_BLK		;Try to allocate a block
	POP	BX			;Restore
	OR	DI,DI			;Success?
	JNZ	LOCK2
;Unable to allocate block
	XOR	AX,AX
	JMP	LOCKX
;We've got one -- swap us in.
LOCK2:
	DEC	DI			;Make byte address
	SHL	DI,1
	SHL	DI,1
	SHL	DI,1
	XCHG	BX,DI
	CALL	SWAPIN			;Swap us in.
	JMP	LOCK0			;Lock

	PUBLIC	_unlockhandle
;
;_unlockhandle
; Call this routine to unlock a memory block previously
; locked.  It has no effect on an already-unlocked block.
; Call with:
; unsigned int unlockhandle(hnum)
; unsigned int hnum;
_unlockhandle:
	PUSH	BP
	MOV	BP,SP
	PUSH	DI
	PUSH	ES
	MOV	AX,FREE_START
	MOV	ES,AX
;Resolve mnode number into byte address
	MOV	DI,[BP+4]
	DEC	DI
	SHL	DI,1
	SHL	DI,1
	SHL	DI,1
;Unlock
	AND	ES:[DI+PREVPTR],7FFFH	;Clear hi bit
	POP	ES
	POP	DI
	POP	BP
	RET
;++++++++++++++++++++++++++++  
; INIT_HEAP
;++++++++++++++++++++++++++++
; entry: AX = # of paragraphs in free memory
;        BX = starting segment of free memory
;        CX = number of mnodes requested
INIT_HEAP:
	PUSH	ES		;Save these guys
	PUSH	DI
	MOV	FREE_START,BX
	MOV	ES,BX
;Calculate the number of bytes used in the mnodes
;region.
	MOV	DX,CX
	SHL	DX,1		;# of mnodes * 8
	SHL	DX,1
	SHL	DX,1
;Calculate the starting segment of the heap data area
	ADD	DX,15
	SHR	DX,1
	SHR	DX,1
	SHR	DX,1
	SHR	DX,1
;Calculate remaining paragraphs
	SUB	AX,DX
	MOV	HEAP_LEFT,AX
	ADD	DX,BX
	MOV	HEAP_START,DX
;Set up initial mnode
	MOV	WORD PTR NUMNODELST,1	;Start of list...
	MOV	WORD PTR NUMNODETOP,1	;...is also end of list.
	XOR	DI,DI
	XCHG	DX,AX
	STOSW			;Base
	MOV	AX,DX
	STOSW			;Length
	XOR	AX,AX
	STOSW			;No next pointer
	STOSW			;No previous pointer
;Build doubly linked-list of mnodes
	MOV	AX,2		;Mnode 2 is start
	MOV	WORD PTR FREEMNODELST,AX
	DEC	CX		;Don't use first pointer
	ADD	DI,4		;Bump to second mnode
ILP1:	INC	AX
	STOSW
	ADD	DI,6
	LOOP	ILP1
	SUB	DI,8		;Fix last mnode -- needs terminator
	XOR	AX,AX
	STOSW
;Restore registers and go home
	POP	DI
	POP	ES
	RET
;++++++++++++++++++++++++++++
; ALLOC_BLK
;++++++++++++++++++++++++++++
; Try to allocate a memory block.
; On entry: DX = # of paragraphs to allocate
;           ES = segment of mnode region
; On exit:  DI = mnode number
;              = 0 if request fails
ALLOC_BLK:
	CALL	SRCHINMEM		;Look in memory first
	DEC	DI
	JS	ALLOCB1
;Got one
	JMP	CLAIM_BLK
;Memory search failed.  Try compacting the heap.
ALLOCB1:
	PUSH	DX
	CALL	COMPACT
	POP	DX
	CMP	CX,DX			;Did we get enough?
	JAE	ALLOC_BLK		;Yes, go get it
;We failed -- can we do a swap?
	CALL	FIND_SWAP_CHAIN		;Find list of blocks to swap
	DEC	DI			;Success?
	JNS	ALLOCB2
;Come here in failure
ALLOCX:
	XOR	DI,DI
	RET
;We found a chain -- set the swap destination and do the swap
ALLOCB2:
	CALL	SETSWAPDEST
	CALL	SWAP_CHAIN
	DEC	DI			;Did it work?
	JS	ALLOCX			;Rats!
;It worked -- go claim the freed block.
;	JMP	CLAIM_BLK
;
;++++++++++++++++++++++++++++
; CLAIM_BLK
;++++++++++++++++++++++++++++
; ALLOC_BLK comes here if a block was found that's big enough to
; fulfill the request.  There's one hitch: the request can still
; fail if the block has more memory than we need, and there are
; no new mnodes to reference the unused portion.
; On entry: DI = byte offset to mnode
;           ES = segment of mnode region
;           DX = # of paragraphs to claim from this block.
; On exit:  DI = mnode number
;           DI = 0 if fail
CLAIM_BLK:
	PUSH	AX
	PUSH	BX
	MOV	AX,ES:[DI+BLKLENGTH]
	SUB	AX,DX		;Grab what you need
	JNZ	CLAIM1		;Jump if anything left
;Nothing left, show block as in use and scram
	OR	WORD PTR ES:[DI+PREVPTR],4000H
CLAIMX:	SHR	DI,1		;Make it an mnode number
	SHR	DI,1
	SHR	DI,1
	INC	DI
CLAIM0:	POP	BX
	POP	AX
	RET
;There's some memory left, request a new mnode to handle the
;overflow and tie this in with the current pointer node.
CLAIM1:
	MOV	BX,DI		;Save current pointer
	CALL	GET_MNODE	;Get a new mnode
	DEC	DI		;Did we get one?
	JZ	CLAIM0		;Arrgh!!
	MOV	ES:[BX+BLKLENGTH],DX	;Adjust length
	ADD	DX,ES:[BX+BLKBASE]	;Get new base
	OR	ES:[BX+PREVPTR],4000H	;Mark as in use
	MOV	ES:[DI+BLKLENGTH],AX	;Set new mnode's length
	MOV	ES:[DI+BLKBASE],DX	;Set new mnode's base
;Link in new mnode after current mnode
	CALL	LINK_AFTERNU
	MOV	DI,BX		;Pass proper handle back
;Return with mnode number in DI
	JMP	CLAIMX
;
;++++++++++++++++++++++++++++
; SRCHINMEM
;++++++++++++++++++++++++++++
; Search the in-memory list for a free block.
; entry: DX = # of paragraphs requested
;        ES = points to mnode segment
; exit: DI = BYTE OFFSET of mnode +1.
;          = 0 if failed
SRCHINMEM:
	MOV	DI,NUMNODELST	;Get head of list
SRCHM1:	DEC	DI		;Make it a true byte offset
	SHL	DI,1
	SHL	DI,1
	SHL	DI,1
;Is the memory block in use?
	TEST	ES:[DI+PREVPTR],4000H
	JNZ	SRCHM2
;Block is not in use -- does it have enough space?
	CMP	DX,ES:[DI+BLKLENGTH]
	JA	SRCHM2
;Enough space! I'll take this one
	INC	DI
	RET
;Block is either in use or too small.  Try next one.
SRCHM2:	MOV	DI,ES:[DI+NEXTPTR]
	AND	DI,3FFFH		;Clear out hi bits
	JNZ	SRCHM1
;We've hit th end of the list! Too bad!
	RET
;
;++++++++++++++++++++++++++++
; COMPACT
;++++++++++++++++++++++++++++
; This is the compaction routine.  It passes through the list of
; in-use m-nodes, forceing "bubbles" of unused memory to the top
; (or to the nearest locked block).
;
COMPACT:
	XOR	AX,AX
	MOV	SI,AX		;SI holds current "bubble"
	MOV	CX,AX		;CX gathers largest bubble size
	MOV	DRTYFLAG,AL	;Indicates a bubble has moved
;Start at lowest m-node
	MOV	DI,NUMNODELST
;Look for next block
CMPCT1:
	OR	DI,DI
	JNZ	CMPCT2
	JMP	CMPCT8
;See if this block is in use
CMPCT2:
	DEC	DI		;Resolve to byte address
	SHL	DI,1
	SHL	DI,1
	SHL	DI,1
	TEST	ES:[DI+PREVPTR],4000H
	JNZ	CMPCT3
	JMP	CMPCT6
;This block is in use, see if it is moveable (unlocked)
CMPCT3:
	TEST	ES:[DI+PREVPTR],8000H
	JNZ	CMPCT5
;We can move this block (if we want).  See if something
;is in SI (we have a bubble to percolate).
	OR	SI,SI
	JZ	CMPCT4
;We do have a bubble, move the block
	MOV	AX,ES:[DI+BLKBASE]	;Source
	MOV	BX,ES:[SI+BLKBASE]	;Destination
	MOV	DX,ES:[DI+BLKLENGTH]	;Length
	PUSH	DX			;Save stuff
	PUSH	DS
	PUSH	ES
	MOV	DS,AX
	MOV	ES,BX
	CALL	copy_down		;Do the move
	POP	ES
	POP	DS
;Adjust the base address in the block
	MOV	ES:[DI+BLKBASE],BX
	POP	DX
;The bubble moves up in memory
	ADD	ES:[SI+BLKBASE],DX
	MOV	BYTE PTR DRTYFLAG,1	;Set dirty flag
;Go on to next m-node
CMPCT4:
	MOV	BX,DI
	MOV	DI,ES:[DI+NEXTPTR]
	JMP	CMPCT1
;The current block is locked.  If the dirty flag is set, then
;the m-node saved in SI has to be "moved".  Do this by deleting
;the m-node, disconnecting it from the in-use list, and creating
;a new one to wire in.
CMPCT5:
	MOV	AL,DRTYFLAG
	OR	AL,AL
	JZ	CMPCT4
	MOV	AX,ES:[SI+BLKBASE]	;Save base
	MOV	BX,ES:[SI+BLKLENGTH]	;Save length
	XCHG	SI,DI
	CALL	RELEASE_MNODE		;Cut it loose
	CALL	GET_MNODE		;Grab a new one
	MOV	ES:[DI+BLKBASE],AX	;Restore base
	MOV	ES:[DI+BLKLENGTH],BX	;Restore length
	MOV	BX,SI			;Current m-node in BX
	CALL	LINK_BEFORENU		;Wire the new m-node in
	XOR	SI,SI			;Clear saved m-node
	MOV	BYTE PTR DRTYFLAG,0	;Clear dirty flag
	MOV	DI,BX			;Current goes into DI
	JMP	CMPCT4			;Move along, now
;The block is not in use.  See if there's anything in SI.  If
;so, add this block to it...the bubble grows.
CMPCT6:
	OR	SI,SI			;Anything there?
	JNZ	CMPCT6A
;Nothing in SI.  Block in DI becomes the new bubble.
	MOV	SI,DI
	MOV	AX,ES:[SI+BLKLENGTH]
	CMP	CX,AX			;New largest?
	JAE	CMPCT4
	MOV	CX,AX
	JMP	CMPCT4
;Add block to current bubble.
CMPCT6A:
	MOV	AX,ES:[SI+BLKLENGTH]	;Get current bubble size
	ADD	AX,ES:[DI+BLKLENGTH]	;New size
	MOV	ES:[DI+BLKLENGTH],AX	;Save
	CMP	CX,AX			;Bigger than biggest?
	JAE	CMPCT7
	MOV	CX,AX			;New winner
CMPCT7:	XCHG	SI,DI
	CALL	RELEASE_MNODE		;Clear the old saved m-node
	MOV	DI,SI
	MOV	BYTE PTR DRTYFLAG,0	;Clear dirty flag
	JMP	CMPCT4
;We've hit the end of the list -- is the dirty bit set?
CMPCT8:
	CMP	BYTE PTR DRTYFLAG,0
	JZ	CMPCTX
;Dirty bit is set.  Release the saved m-node and wire a
;new one in at the end of the list.
	MOV	AX,ES:[SI+BLKBASE]  	;Save base
	MOV	BX,ES:[SI+BLKLENGTH]	;Save length
	XCHG	DI,SI
	CALL	RELEASE_MNODE
	CALL	GET_MNODE
	MOV	ES:[DI+BLKBASE],AX	;Set base
	MOV	ES:[DI+BLKLENGTH],BX	;Set length
	MOV	BX,SI
	CALL	LINK_AFTERNU		;Link after current
CMPCTX:	RET
;++++++++++++++++++++++++++++
; SWAP_CHAIN
;++++++++++++++++++++++++++++
; On entry, DI holds the byte address of a starting m-node.
; This routine moves upwards along the m-node list, swapping
; used blocks out to disk.  The routine halts when it encounters
; (a) a locked block, (b) the end of the list, or (c) a free block.
; All swapped-out areas are consolidated into a single free block.
; The routine returns the m-node byte address +1 in DI, or 0 if
; the routine fails.
SWAP_CHAIN:
	MOV	SI,DI			;Save start of chain
	CALL	GET_MNODE		;Try to get a new mnode
	DEC	DI			;One found?
	JNS	SWAPCH1
;Whoops! No free m-nodes
	INC	DI
	RET
;Put new m-node in SI and start working.  We'll use this
;m-node to point to the block that will be freed when
;the swap-chain is gone.
SWAPCH1:
	XCHG	SI,DI
	MOV	AX,ES:[DI+BLKBASE]
	MOV	ES:[SI+BLKBASE],AX		;Set base
	MOV	CX,ES:[DI+PREVPTR]
SWAPCH2:
	MOV	AX,ES:[DI+BLKLENGTH]
	ADD	ES:[SI+BLKLENGTH],AX	;Accumulate length
	MOV	BX,ES:[DI+NEXTPTR]	;Get next m-node in chain
	CALL	UNLINKNU		;Unlink the m-node
	TEST	CH,0C0H			;Free block?
	JZ	SWAPCH5			;No swap if so
	CALL	SWAPOUT			;Swap out current m-node
;See what next m-node is
	AND	BX,3FFFH		;Clear location bits
	JZ	SWAPCH6			;At end of list!
	DEC	BX 			;Resolve to byte offset
	SHL	BX,1
	SHL	BX,1
	SHL	BX,1
	MOV	DI,BX
	MOV	CX,ES:[DI+PREVPTR]	;Get status bits
	TEST	CH,80H			;Locked?
	JZ      SWAPCH2			;Keep going if not
;We've hit a locked block.  Wire the node in SI in prior
;to this fellow.
	MOV	DI,SI			;New block in DI
	CALL	LINK_BEFORENU		;Link block in list
	INC	DI
	RET				;Go home
;Come here if we hit the top part of a bubble.  Put this member
;onto the free list.  Then link in the new node.
SWAPCH5:
	CALL	LINK_BEFOREF		;Put on free list
SWAPCH6:
	MOV     BX,CX			;Set current
	DEC	BX
	SHL	BX,1
	SHL	BX,1
	SHL	BX,1			;Make it a byte address
	MOV	DI,SI			;Set new node
	CALL	LINK_AFTERNU		;Link in new bubble
	INC	DI
	RET
;
;++++++++++++++++++++++++++++
; FIND_SWAP_CHAIN
;++++++++++++++++++++++++++++
; This routine searches down through the mnode list, looking
; for a sequence of unlocked blocks that can be swapped out to
; satisfy requrest for # of paragraphs in DX.
; On success, DI holds byte address+1 of starting mnode.
; BX holds # of paragraphs that will need to be swapped.
; On failure, DI holds 0.
FIND_SWAP_CHAIN:
	MOV	DI,NUMNODETOP		;Start at top of list
	XOR	CX,CX			;Clear length
	XOR	BX,BX			;Clear # swappable paragraphs
;Calc. m-node's byte address
FSWCH1:
	DEC	DI
	SHL	DI,1
	SHL	DI,1
	SHL	DI,1
;See if block is locked, if so, start from scratch
	MOV	AX,ES:[DI+PREVPTR]	;Get previous pointer
	TEST	AH,80H
	JZ	FSWCH2
	XOR	CX,CX			;Block is locked
	XOR	BX,BX
	JMP	FSWCH3
;Not locked, increment length
FSWCH2:
	TEST	AH,0C0H
	JZ	FSWCH3			;Don't count free blocks
	ADD	BX,ES:[DI+BLKLENGTH]
FSWCH3:
	ADD	CX,ES:[DI+BLKLENGTH]
	CMP	DX,CX			;Got enough?
	JA	FSWCH4			;Jump if not
;We have enough, go home
	INC	DI			;Show ok
	RET
;Keep looking through list
FSWCH4:
	MOV	DI,AX
	AND	DI,3FFFH
	JNZ	FSWCH1
	RET				;We've hit the list end
;
;++++++++++++++++++++++++++++
; ** SWAPIN **
;++++++++++++++++++++++++++++
; Swap a block in from somewhere.
; BX = byte address of mnode to claim memory from.
; DI = byte address of mnode to swap in.
SWAPIN:
	MOV	DX,ES:[BX+BLKBASE]	;Segment
	MOV	CX,ES:[DI+BLKLENGTH]	;# of paragraphs
	PUSH	CX
	PUSH	BX
	MOV	BX,ES:[DI+BLKBASE]	;Disk block/MAT
	MOV	AX,ES:[DI+NEXTPTR]	;Where from?
	AND	AH,0C0H
	CMP	AH,80H
	JZ	SWAPIN1
;Comes from EMS
	CALL	READFROMEMS
	JMP	SWAPIN2
;Comes from disk
SWAPIN1:
	CALL	READFROMDISK
;Bring mnode back to reality -- update base address
SWAPIN2:
	POP	BX
	MOV	AX,ES:[BX+BLKBASE]
	MOV	ES:[DI+BLKBASE],AX
;Show that we're in conventional memory
	AND	ES:[DI+NEXTPTR],03FFFH
;Update base address of "next" mnode
	POP	CX		;Restore length
	ADD	ES:[BX+BLKBASE],CX
;Wire us in prior to mnode we stole from
	CALL	LINK_BEFORENU
;Decrement his length
	SUB	ES:[BX+BLKLENGTH],CX
	JNZ	SWAPINX
;He's all use up...release him
	XCHG	BX,DI
	CALL	RELEASE_MNODE
	XCHG	BX,DI
SWAPINX:
	RET
;
;++++++++++++++++++++++++++++
; **SWAPOUT**
;++++++++++++++++++++++++++++
; Swap out an mnode.
; DI holds byte address of mnode to swap
SWAPOUT:
	PUSH	BX
	PUSH	CX
	MOV	BX,ES:[DI+BLKBASE]	;Get seg address
	MOV	CX,ES:[DI+BLKLENGTH]	;Get # of paragraphs
	CMP	BYTE PTR SWAPDEST,1	;Swap to where?
	JNZ	SWAPO2
;Swap this fellow to EMS
	OR	WORD PTR ES:[DI+NEXTPTR],4000H
	CALL	WRITETOEMS
SWAPO1:	MOV	ES:[DI+BLKBASE],AX	;Show where he is
	POP	CX
	POP	BX
	RET
;Swap this guy to disk.
SWAPO2:	OR	WORD PTR ES:[DI+NEXTPTR],8000H
	CALL	WRITETODISK
	JMP	SWAPO1
;
;++++++++++++++++++++++++++++
; ** SETSWAPDEST **
;++++++++++++++++++++++++++++
; Set the swap destination.
; Look for EMS first...if it fails -- go for the disk.
SETSWAPDEST:
	MOV	AL,EMSTHERE		;Do we have EMS?
	OR	AL,AL
	JNZ	SSWAPD1
;Go to disk
SETSWAPTODISK:
	MOV	BYTE PTR SWAPDEST,2
	RET
;EMS present -- enough memory?
SSWAPD1:
	PUSH	DX
	MOV	AX,EMSBLKSFREE
	MOV	CX,16
	MUL	CX
	OR	DX,DX
	POP	DX
	JZ	SSWAPD2
;If DX nonzero - no problem
SETSWAPTOEMS:
	MOV	BYTE PTR SWAPDEST,1
	RET
;Check AX against BX
SSWAPD2:
	CMP	AX,BX
	JA	SETSWAPTOEMS
	JMP	SETSWAPTODISK

;****************************
;   FREEING BLOCKS          *
;****************************
;++++++++++++++++++++++++++++
; FREE_BLK
;++++++++++++++++++++++++++++
; Free a block.  DI is handle to mnode of block to free.
; Returns DI<> 0 if ok, else DI = 0.
;
FREE_BLK:
	DEC	DI		;Resolve handle to address
	SHL	DI,1
	SHL	DI,1
	SHL	DI,1
;See if the block is locked
	MOV	AX,ES:[DI+PREVPTR]
	AND	AH,80H		;Check hi bit
	JNZ	FREEBLK1
;Block is locked - no can do
	XOR	DI,DI
	RET
;Block is not locked - determine its location
FREEBLK1:
	MOV	AX,ES:[DI+NEXTPTR]
	AND	AH,0C0H
	JNZ	FREEBLK2
;Block is in memory.  Mark this m-node as being free now.
	AND	ES:[DI+PREVPTR],03FFFH	;Clear hi two bits
	INC	DI			;Set DI <> 0
	RET
;Get base pointer in BX...this will either be the MAT or disk block#
FREEBLK2:
	MOV	BX,ES:[DI+BLKBASE]
	CMP	AH,4
	JNZ	FREEBLK3
;MAT entry
	CALL	FREE_MATLIST
	CALL	UNLINKNU		;Free the mnode
	INC	DI			;Show success
	RET
;Disk block
FREEBLK3:
	CALL	FREE_DISKBLKS
	CALL	UNLINKNU		;Free the mnode
	INC	DI			;Show success
	RET
;
;++++++++++++++++++++++++++++
;   FREE_MATLIST
;++++++++++++++++++++++++++++
; Free a chain of MAT entries.
; BX holds head of list
FREE_MATLIST:
	CALL	RELEASE_MAT		;Free the head entry
	MOV	BX,NEXTBPTR		;Are there more?
	OR	BX,BX
	JNZ	FREE_MATLIST
	RET
;
; Free a chain of disk blocks
; BX holds head of the chain.
FREE_DISKBLKS:
	CALL	RELEASE_DBL		;Release a block
	MOV	BX,NEXTBPTR		;Are there more?
	OR	BX,BX
	JNZ	FREE_DISKBLKS
	RET
;****************************
;  SWAP FILE ROUTINES       *
;****************************
;
;++++++++++++++++++++++++++++
; OPENSWAP
;++++++++++++++++++++++++++++
; Open the swap file.
; Assumes offset to swapfile name is in DX.
; Returns handle in SWAPHANDLE
; If the file is not found in the local directory, it is created.
; On return, if AX==0, all went well.  Otherwise, AX holds the error.
OPENSWAP:
	PUSH	CX
	MOV	AL,2			;Read/write access
	MOV	AH,3DH			;Function code
	INT	21H			;Call DOS
	JC	OPENSWAPERR
OPENSWAP0:
	MOV	SWAPHANDLE,AX		;Save handle
	XOR	AX,AX			;Show all ok
OPENSWAP1:
	POP	CX
	RET
;An error has occured.  If it is file not found, attempt to create
;the file.
OPENSWAPERR:
	CMP	AX,2			;File not found?
	JZ	OPENSWAPERR1
	CMP	AX,4
	JNZ	OPENSWAP1
OPENSWAPERR1:
	XOR	CL,CL			;Set file attributes
	MOV	AH,3CH			;Create function
	INT	21H
	JC	OPENSWAP1		;I give up
	JNC	OPENSWAP0		;All is well
;
;++++++++++++++++++++++++++++
;  MOVESWAPNAME
;++++++++++++++++++++++++++++
; Move the swap file's name into our local memory
; (We need it so we can delete the file when the
;  program terminates.)
MOVESWAPNAME:
	PUSH	SI
	PUSH	DI
	MOV	SI,DX			;Get pointer to name
	MOV	DI,OFFSET SWAPNAME
;Loop and move the name, byte at a time
MOVSWAPN1:
	LODSB				;Get a character
	MOV	DS:[DI],AL		;Store it
	INC	DI			;Bump DI
	OR	AL,AL			;Null?
	JNZ	MOVSWAPN1
;Go home
	POP	DI
	POP	SI
	RET
;
;++++++++++++++++++++++++++++
;  READFROMDISK
;++++++++++++++++++++++++++++
; Read a memory image in from the disk
; BX = starting block number
; CX = # of paragraphs to read
; DX = segment to load block into
READFROMDISK:
	PUSH	SI		;We'll borrow these registers
	PUSH	DI
;Calculate top of segment
	MOV	DI,DX
	ADD	DI,CX
;Release first block.  The RELEASE routine will leave the file
;pointer at the data area & our next block in NEXTBPTR.
RFDISK1:
	CALL	RELEASE_DBL
;Calculate # of paragraphs by ANDing remaining paragraphs by 3fH.
;If remainder is 0, read 64 paragraphs, else read remainder.
	PUSH	CX
	AND	CX,3FH
	JNZ	RFDISK2
	MOV	CX,64
RFDISK2:
	PUSH	CX		;Push # of paragraphs read
;Adjust destination segment
	SUB	DI,CX
	MOV	BX,SWAPHANDLE
	PUSH	DS
	MOV	DS,DI
	XOR	DX,DX		;Destination offset
	SHL	CX,1		;Resolve CX into # of bytes
	SHL	CX,1
	SHL	CX,1
	SHL	CX,1
;Read the data
	MOV	AH,3FH
	INT	21H
	POP	DS
;See if we're done
	MOV	BX,NEXTBPTR
	POP	AX
	POP	CX
	SUB	CX,AX		;Calc remaining paragraphs
	JNZ	RFDISK1
;All done - restore and return
	POP	DI
	POP	SI
	RET
;
;++++++++++++++++++++++++++++
;  WRITETODISK
;++++++++++++++++++++++++++++
; Write a memory region into the swap file.
;  BX = Segment start of region to write
;  CX = # of paragraphs
; Return with the pointer to the head of the disk block list in AX
WRITETODISK:
	PUSH	DX		;Save pointers
	PUSH	SI
	MOV	SI,BX
;Clear NEXTBPTR
	XOR	AX,AX
	MOV	NEXTBPTR,AX
;Allocate a new block
WTDISK1:
	CALL	ALLOC_DISKB
	MOV	NEXTBPTR,BX
;# of paragraphs to write is 64 and # of
;NOTE: THIS CODE PRESUMES THE SEGMENT WILL SIMPLY WRAP
;IF THIS READ IS NEAR THE END OF THE SEGMENT.  THIS COULD
;CAUSE PROBLEMS WITH DOS EXTENDERS.
	PUSH	CX
	MOV	CX,1024		;64 paragraphs = 1024 bytes
;Set up segment
	MOV	BX,SWAPHANDLE	;Get handle before DS changes
	PUSH	DS
	MOV	DS,SI
	ADD	SI,64		;Ready SI for next tim
;Calculate # of bytes to write
	XOR	DX,DX		;Start of segment
	MOV	AH,40H
	INT	21H		;Write memory block
;See if we're done
	POP	DS		;Restore segment
	POP	CX
	SUB	CX,64
	JZ	WTDISK2
	JNS	WTDISK1
;We are done, exit with block head (ha!) in AX
WTDISK2:
	MOV	AX,NEXTBPTR
	POP	SI
	POP	DX
	RET
;
;++++++++++++++++++++++++++++
;  ALLOC_DISKB
;++++++++++++++++++++++++++++
; Allocate a new block for the swap file
; First attempt to pull the block off the avail list.  If the
; list is empty, extend the file.
; Returns block number in BX.  Also, writes out NEXTBPTR in block
; pointer and leaves file pointer ready at data section.
ALLOC_DISKB:
	PUSH	CX		;Save registers
	PUSH	BX
	PUSH	SI
;See if avail list has anything
	MOV	BX,DISKAVL
	OR	BX,BX
	JZ	ALLOCD2
;Something on avail list -- use it.
	MOV	SI,BX		;Save block #
	CALL	DISKBLOCKTOBOFF
	MOV	AH,42H
	MOV	BX,SWAPHANDLE
	INT	21H		;Seek to block
;Read the block's next link into DSKAVL
	MOV	CX,2
	MOV	DX,OFFSET DISKAVL
	MOV	AH,3FH
	INT	21H		;Read next link in chain
;Back up 2 bytes in file
	MOV	CX,-1
	MOV	DX,-2
	MOV	AL,1
	MOV	AH,42H
	INT	21H
;Write out NEXTBPTR
ALLOCD1:
	MOV	CX,2
	MOV	DX,OFFSET NEXTBPTR
	MOV	AH,40H
	INT	21H
;Restore save block # and exit
	MOV	BX,SI
	POP	SI
	POP	DX
	POP	CX
	RET
;Come here if nothing on avail list.  Get last offset in file.
ALLOCD2:
	XOR	CX,CX
	XOR	DX,DX
	MOV	AL,2
	MOV	AH,42H
	MOV	BX,SWAPHANDLE
	INT	21H		;Position to EOF
	MOV	CX,1026
	DIV	CX		;Calculate block #
	INC	AX		;0th block is # 1
	MOV	SI,AX		;Save block number
	JMP	ALLOCD1		;Exit by writing out NEXTBPTR
;
;++++++++++++++++++++++++++++
;  RELEASE_DBL
;++++++++++++++++++++++++++++
; Release a block from the SWAP file.
; BX = block number to release.
; This routine leaves the file pointer positioned at the start
; of the data region for this block.  It also leave's the
; block's NEXT pointer in NEXTBPTR.
RELEASE_DBL:
	PUSH	CX		;Save registers we'll cream
	PUSH	DX
;Save incoming block number
	PUSH	BX
;Seek to released block
	CALL	DISKBLOCKTOBOFF
	MOV	BX,SWAPHANDLE
	MOV	AL,0
	MOV	AH,42H
	INT	21H		;Seek
;Read this block's NEXT link
	MOV	CX,2		;2 bytes
	MOV	DX,OFFSET NEXTBPTR
	MOV	AH,3FH
	INT	21H
;Back up 2 bytes
	MOV	CX,-1
	MOV	DX,-2
	MOV	AL,1		;Seek from current
	MOV	AH,42H
	INT	21H
;Write the current AVAIL list head
	MOV	CX,2
	MOV	DX,OFFSET DISKAVL
	MOV	AH,40H
	INT	21H
;Saved block becomes new AVAIL list head
	POP	BX
	MOV	DISKAVL,BX
	POP	DX
	POP	CX
	RET
;
;++++++++++++++++++++++++++++
;  DISKBLOCKTOBOFF
;++++++++++++++++++++++++++++
; Given a disk block # in BX, return the corresponding
; byte offset in CX:DX (ready for an lseek).
DISKBLOCKTOBOFF:
	DEC	BX
	PUSH	AX
	MOV	AX,1026
	MUL	BX
	MOV	CX,DX
	MOV	DX,AX
	POP	AX
	RET
;****************************
;    EMS ROUTINES           *
;****************************
;
;++++++++++++++++++++++++++++
;  ISERMSHERE
;++++++++++++++++++++++++++++
; Verify that EMS is available.  Returns 0 in AX if EMS is
; not available. Otherwise, returns number of unallocated
; EMS pages.
ISEMSHERE:
	PUSH	BX		;Save these guys
	PUSH	DX
	PUSH	SI
	MOV	AL,67H		;EMS interrupt vector
	MOV	AH,35H		;Get interrupt vector
	INT	21H
	MOV	DI,0AH		;Point to device name string
	LEA	SI,EMSDEVNAM
	MOV	CX,EMSDEVNAMLEN
	CLD			;Increment
	REPE	CMPSB		;Compare
	JZ	ISEMSHERE1
	;EMS not here, return 0
EMSNOTHERE:
	XOR	AX,AX
	POP	DX
	POP	BX
	RET
	;Find out how many free pages we've got.
ISEMSHERE1:
	MOV	AH,42H		;Get unallocated pages
	INT	67H
	OR	AH,AH
	JNZ	EMSNOTHERE	;If error, treat as not there
	MOV	AX,BX
	POP	SI
	POP	DX
	POP	BX
	RET
;
;++++++++++++++++++++++++++++
;  GETPFSEGADDR
;++++++++++++++++++++++++++++
;Get the EMS page frame segment address
;Returns 0 in AX if ok, else error code.
;
GETPFSEGADDR:
	PUSH	BX
	MOV	AH,41H
	INT	67H
	OR	AH,AH
	JNZ	GETPFSX
	MOV	EMSPFSEG,BX	;Save the segment
GETPFSX:
	XOR	AL,AL
	POP	BX
	RET
;
;++++++++++++++++++++++++++++
;  INIT_EMS
;++++++++++++++++++++++++++++
;Initialize the EMS avail list.
;Call with AX=# of 16K pages to grab.
;Returns AX=0 if all went well, else returns error code.
INIT_EMS:
	PUSH	BX	;We'll need most registers
	PUSH	CX
	PUSH	DX
	PUSH	SI
	PUSH	DI
;Attempt to allocate pages
	MOV	BX,AX		;# pages in BX
	CALL	GETPFSEGADDR	;Get page frame seg. address
	OR	AH,AH
	JNZ	INITEMSX	
	MOV	CX,BX		;Save # pages in CX for later
	MOV	AH,43H
	INT	67H
;Did allocation work?
	OR	AH,AH
	JZ	INITEMS1
;Allocation failed...exit
INITEMSX:
	POP	DI
	POP	SI
	POP	DX
	POP	CX
	POP	BX
	RET
;Allocation succeeded...save handle
INITEMS1:
	MOV	MYEMSHANDLE,DX
;Save # of free EMS blocks.
	MOV	AX,CX
	MOV	BX,126
	MUL	BX
	MOV	EMSBLKSFREE,AX
;Build the avail list.
;Map in page.
	XOR	BX,BX		;Start with first (0th) page
	MOV	SI,0400H	;For avail list.
	MOV	EMSAVL,SI
INITEMS2:
	MOV	AX,4400H
	MOV	DX,MYEMSHANDLE
	INT	67H
	OR	AH,AH
	JNE	INITEMSX	;What on earth?
;Build MAT list for this page.
	PUSH	CX		;Save counter
	MOV	DI,4		;Starting guy's addr
	MOV	CX,125		;Loop will handle 125 MAT entries
	PUSH	ES
	MOV	DX,EMSPFSEG	;Switch to PF segment
	MOV	ES,DX
	MOV	AX,SI
INITEMS3:
	ADD	AH,2		;Increment MAT offset
	STOSW			;Store MAT entry
	LOOP	INITEMS3
;Last MAT points across the page to first MAT of next page.
	INC	BX
	MOV	AX,BX
	OR	AH,4		;First entry is element 2
	STOSW
	POP	ES
	MOV	SI,AX
;Keep going
	POP	CX
	LOOP	INITEMS2
;Very last MAT entry must be all 0's to indicate end of list.
	MOV	CX,ES
	MOV	ES,DX
	SUB	DI,2
	XOR	AX,AX
	STOSW
	MOV	ES,CX
;All's well, go home
	JMP	INITEMSX

;
;++++++++++++++++++++++++++++
;  CLAIM_MAT
;++++++++++++++++++++++++++++
; Claim a MAT entry.
; Returns with claimed MAT entry in AX.
; AX = 0 if failure.
; *** NOTE: MODIFY THIS GUY TO DECREMENT # OF FREE MAT ENTRIES***
CLAIM_MAT:
	PUSH	BX		;Save things
	PUSH	CX
	PUSH	DX
	MOV	AX,EMSAVL	;Anything there?
	OR	AX,AX
	JNZ	CLMMAT1
;Failure...it's so sad.
CLMMATX:
	XOR	AX,AX
	POP	DX
	POP	CX
	POP	BX
	RET
;Bank in the page
CLMMAT1:
	MOV	AX,4400H	;Function and physical page
	MOV	BX,EMSAVL
	AND	BH,01H		;Clear MAT offset bits
	MOV	DX,MYEMSHANDLE
	INT	67H
	OR	AH,AH		;All ok?
	JNZ	CLMMATX		;Jump if not
;Unlink from EMS avail list
	MOV	CX,EMSAVL	;Save in CX
	MOV	BL,CH
	AND	BX,00FEH	;Mat offset in BX
	PUSH	ES
	MOV	AX,EMSPFSEG
	MOV	ES,AX
	MOV	AX,ES:[BX]	;Get next MAT entry
	POP	ES
	MOV	EMSAVL,AX
	MOV	AX,CX
	POP	DX
	POP	CX
	POP	BX
	RET
;
;++++++++++++++++++++++++++++
;  RELEASE_MAT
;++++++++++++++++++++++++++++
; Release a MAT entry.
; BX = MAT to release
; Leave's MAT's next link in NEXTBPTR
RELEASE_MAT:
	PUSH	BX
	PUSH	CX
	PUSH	DX
	PUSH	DI
;Save avail ptr
	MOV	DX,EMSAVL
	PUSH	DX
;New MAT goes on fron of avail
	MOV	EMSAVL,BX
;Bank in the released MAT's page.
	MOV	CL,BH
	MOV	AX,4400H
	AND	BX,1FFH
	MOV	DX,MYEMSHANDLE
	INT	67H
;**CHECK FOR NONRECOVERABLE ERROR SHOULD GO HERE**
	MOV	DI,CX	
	AND	DI,0FEH		;Mat offset in DX
	POP	AX
	PUSH	ES
	MOV	BX,EMSPFSEG
	MOV	ES,BX
	MOV	BX,ES:[DI]	;Get next link
	STOSW
	POP	ES
	MOV	NEXTBPTR,BX
;All's well, go home.
	POP	DI
	POP	DX
	POP	CX
	POP	BX
	RET
;
;++++++++++++++++++++++++++++
;  WRITETOEMS
;++++++++++++++++++++++++++++
; Write a block from the heap to EMS.
; Call with BX = segment of region to write.
;           CX = # of paragraphs
; Return with AX = MAT entry of head of list.
WRITETOEMS:
	PUSH	DX		;Save registers
	PUSH	SI
	PUSH	DI
	XOR	AX,AX		;Setup list tail.
	MOV	NEXTBPTR,AX
;Allocate a new MAT and record the list pointers
WRITETOEMS1:
	CALL	CLAIM_MAT
	MOV	SI,AX		;Save the MAT entry
	MOV	AL,AH
	AND	AX,0FEH
	MOV	DI,AX		;Address of MAT entry in DI
	MOV	AX,NEXTBPTR
	PUSH	ES
	MOV	DX,EMSPFSEG
	MOV	ES,DX
	MOV	ES:[DI],AX	;Build the links
	POP	ES
	MOV	NEXTBPTR,SI	;Save next link
	MOV	DI,SI		;Restore MAT
;Time to write the memory.  For each pass, the number
;of paragraphs to write is the smaller of 8 and the # of
;paragraphs remaining.
	PUSH	CX
	CMP	CX,8
	JLE	WRITETOEMS2
	MOV	CX,8
WRITETOEMS2:
;Resolve CX into # of words
	PUSH	CX
	SHL	CX,1
	SHL	CX,1
	SHL	CX,1
;Set up source and destination offsets
	XOR	SI,SI		;Source is easy
	AND	DI,0FE00H	;Calculate mat's assoc. block addr.
	SHR	DI,1
	SHR	DI,1
;Set up segments
	PUSH	DS
	PUSH	ES
	MOV	AX,EMSPFSEG
	MOV	ES,AX
	MOV	DS,BX
	CLD
	REP	MOVSW
	POP	ES
	POP	DS
;Update source segment
	POP	AX		;# paragraphs moved
	ADD	BX,AX
;See if we're done
	POP	CX		;# of paragraphs remaining
	SUB	CX,AX
	JNZ	WRITETOEMS1	;Keep going
;Go home with MAT pointer
	 MOV	AX,NEXTBPTR
	 POP	DI
	 POP	SI
	 POP	DX
	 RET
;
;++++++++++++++++++++++++++++
;  READFROMEMS
;++++++++++++++++++++++++++++
; Read from EMS
; BX = MAT anchoring chain.
; CX = # paragraphs
; DX = starting segment to read into
READFROMEMS:
	PUSH	SI
	PUSH	DI
;Calculate the top of our destination segment
	ADD	DX,CX
;Release the current MAT.  The RELEASE_MAT function will bank
;in the EMS for us.
RFEMS1:
	CALL	RELEASE_MAT
;Determine # of paragraphs to copy out.  Do this by ANDing # paragraphs
;remaining with 7.  If resule it 0, copy out 8.  Else copy out result.
	PUSH	CX		;Save # of remaining paragraphs
	AND	CX,7
	OR	CX,CX
	JNZ	RFEMS2
	MOV	CX,8
RFEMS2:
	PUSH	CX
;Adjust the destination segment
	SUB	DX,CX
	XOR	DI,DI		;Set up destination offset
;Resolve CX into # of words to move
	SHL	CX,1
	SHL	CX,1
	SHL	CX,1
;Calculate source offset
	AND	BX,0FE00H
	MOV	SI,BX
	SHR	SI,1
	SHR	SI,1
;Move the memory
	PUSH	DS
	PUSH	ES
	MOV	AX,EMSPFSEG
	MOV	DS,AX
	MOV	ES,DX
	CLD
	REP	MOVSW
	POP	ES
	POP	DS
;Get the next MAT
	MOV	BX,NEXTBPTR
;See if we're done
	POP	AX
	POP	CX
	SUB	CX,AX
	JNZ	RFEMS1
;We are finished
	POP	DI
	POP	SI
	RET
;****************************
;  LOWLEVEL MNODE HANDLING  *
;****************************
;++++++++++++++++++++++++++++
;  LINK_BEFORENU
;++++++++++++++++++++++++++++
; Link a new m-node into the in-use list before the current
; m-node.
; On entry:
;  BX = current m-node's byte address
;  DI = new m-node's byte address
LINK_BEFORENU:
	CALL	LINK_BEFORE		;Link it in
	TEST	WORD PTR ES:[DI+PREVPTR],3FFFH
	JNZ	LINKBFNUX
	PUSH	DI
	SHR	DI,1
	SHR	DI,1
	SHR	DI,1
	INC	DI
	MOV	NUMNODELST,DI		;New head of list
	POP	DI
LINKBFNUX:
	RET
;
;++++++++++++++++++++++++++++
; GET_MNODE
;++++++++++++++++++++++++++++
; Get an m-node from the free list.
; Return byte address+1 of mnode in DI
GET_MNODE:
	PUSH	AX
	MOV	DI,FREEMNODELST
	OR	DI,DI			;Anything there?
	JNZ	GETM1			;Jump if so
	RET				;Else go home
;Unlink the m-node from the free list
GETM1:	DEC	DI
	SHL	DI,1
	SHL	DI,1			;Byte address
	SHL	DI,1
	PUSH	DI
	MOV	DI,ES:[DI+NEXTPTR]
	AND	DI,3FFFH
	MOV	FREEMNODELST,DI		;Set new list head
; Clear prev pointer (set when we attached )
	DEC	DI
	JS	GETM2
	SHL	DI,1
	SHL	DI,1
	SHL	DI,1
	MOV	WORD PTR ES:[DI+PREVPTR],0
GETM2:	POP	DI
	XOR	AX,AX
	STOSW				;Clear it out
	STOSW
	STOSW
	STOSW
	SUB	DI,7			;byte address+1
	POP	AX
	RET
;
;++++++++++++++++++++++++++++
;  RELEASE_MNODE
;++++++++++++++++++++++++++++
; Release an m-node.
; DI holds byte pointer to m-node to release.
RELEASE_MNODE:
	CALL	UNLINKNU		;Unlink from in-use list
;	JMP	LINK_BEFOREF		;Put on free list
;
; Link a deleted m-node into the free list.
; Call format same as LINK_BEFORENU, except nothing is needed
; in BX.
LINK_BEFOREF:
	MOV	WORD PTR ES:[DI+PREVPTR],0
	MOV	WORD PTR ES:[DI+NEXTPTR],0
	PUSH	BX
	MOV	BX,FREEMNODELST
	OR	BX,BX
	JZ	LBFF1
	DEC	BX
	SHL	BX,1
	SHL	BX,1
	SHL	BX,1			;Byte address in BX
	CALL	LINK_BEFORE		;Link it in
LBFF1:
	PUSH	DI
	SHR	DI,1
	SHR	DI,1
	SHR	DI,1
	INC	DI
	MOV	FREEMNODELST,DI		;New head of list
	POP	DI
LINKBFFX:
	POP	BX
	RET
;
;++++++++++++++++++++++++++++
;  LINK_BEFORE
;++++++++++++++++++++++++++++
; Link a new m-node into a list **before** the current m-node.
; On entry:
;  BX = current m-nodes's byte address
;  DI = new m-node's byte address
LINK_BEFORE:
	PUSH	AX
	PUSH	CX
	PUSH	DX
	PUSH	DI
; Load DX with an oft-used constant
	MOV	DX,0C000H
; Fix previous links
	MOV	AX,ES:[BX+PREVPTR]
	AND	AX,03FFFH
	PUSH	AX
	AND	ES:[DI+PREVPTR],DX
	OR	ES:[DI+PREVPTR],AX
	MOV	AX,DI
	SHR	AX,1
	SHR	AX,1
	SHR	AX,1
	INC	AX
	MOV	CX,AX
	AND	ES:[BX+PREVPTR],DX
	OR	ES:[BX+PREVPTR],AX
;Fix next links
	MOV	AX,BX
	SHR	AX,1
	SHR	AX,1
	SHR	AX,1
	INC	AX
	AND	ES:[DI+NEXTPTR],DX
	OR	ES:[DI+NEXTPTR],AX
	POP	DI
	OR	DI,DI		;Predecessor?
	JZ	LINBX		;Bail out if not
;Fix predecessor's next link
	DEC	DI
	SHL	DI,1
	SHL	DI,1
	SHL	DI,1
	AND	ES:[DI+NEXTPTR],DX
	OR	ES:[DI+NEXTPTR],CX
LINBX:	POP	DI
	POP	DX
	POP	CX
	POP	AX
	RET
;
;++++++++++++++++++++++++++++
;  LINK_AFTERNU
;++++++++++++++++++++++++++++
; Link in a new m-node after the current m-node in the in-use list.
; Entry:
;	BX = current m-node's byte address
;	DI = new m-node's byte address
LINK_AFTERNU:
	PUSH	AX
	PUSH	CX
	PUSH	DX
	PUSH	DI
;Borrow DX to hold mask
	MOV	DX,0C000H
;Fix next links
	MOV	AX,ES:[BX+NEXTPTR]
	AND	AX,3FFFH
	PUSH	AX
	AND	ES:[DI+NEXTPTR],DX
	OR	ES:[DI+NEXTPTR],AX
	MOV	AX,DI
	SHR	AX,1
	SHR	AX,1
	SHR	AX,1
	INC	AX
	MOV	CX,AX
	AND	ES:[BX+NEXTPTR],DX
	OR	ES:[BX+NEXTPTR],AX
;Fix previous links
	MOV	AX,BX
	SHR	AX,1
	SHR	AX,1
	SHR	AX,1
	INC	AX
	AND	ES:[DI+PREVPTR],DX
	OR	ES:[DI+PREVPTR],AX
	POP	DI
	OR	DI,DI	;Is there a successor node?
	JNZ	LINA1
;If there is no successor, new mnode is new top of list
	MOV	NUMNODETOP,CX
	JMP	LINAX
;Fix successor's previous link
LINA1:
	DEC	DI
	SHL	DI,1
	SHL	DI,1
	SHL	DI,1
	AND	ES:[DI+PREVPTR],DX
	OR	ES:[DI+PREVPTR],CX
LINAX:	POP	DI
	POP	DX
	POP	CX
	POP	AX
	RET
;
;++++++++++++++++++++++++++++
;  UNLINKNU
;++++++++++++++++++++++++++++
; Unlink an mnode from the in-use list.
; Call with DI=byte offset to mnode.
UNLINKNU:
	PUSH	BX		;Save stuff
	PUSH	CX
	PUSH	DX
	PUSH	SI
	PUSH	DI
;Borrow SI to hold mask
	MOV	SI,0C000H
	MOV	BX,ES:[DI+PREVPTR]
	MOV	DI,ES:[DI+NEXTPTR]
	AND	BH,3FH
	AND	DI,3FFFH
	MOV	CX,BX
	MOV	DX,DI
;Fix previous mnode's next link
	OR	BX,BX		;Anything there?
	JNZ	RLS1
	MOV	NUMNODELST,DX	;Nothing there, next is new head
	JMP	RLS2
RLS1:	DEC	BX
	SHL	BX,1
	SHL	BX,1
	SHL	BX,1
	AND	ES:[BX+NEXTPTR],SI
	OR	ES:[BX+NEXTPTR],DX
;Fix next mnode's previous link
RLS2:	OR	DI,DI
	JNZ	RLS3
	MOV	NUMNODETOP,CX	;Nothing there, previous is new top
	JMP	RLS4
RLS3:	DEC	DI
	SHL	DI,1
	SHL	DI,1
	SHL	DI,1
	AND	ES:[DI+PREVPTR],SI
	OR	ES:[DI+PREVPTR],CX
RLS4:	POP	DI
	POP	SI
	POP	DX
	POP	CX
	POP	BX
	RET
;
;++++++++++++++++++++++++++++
;  COPY_DOWN
;++++++++++++++++++++++++++++
;Copy a block of memory downwards.
; DS = source segment base
; ES = destination segment base
; DX = # of paragraphs to move
; We presume DS>ES
COPY_DOWN:
	PUSH	AX
	PUSH	BX
	PUSH	CX
	PUSH	SI
	PUSH	DI
	MOV	BX,1000H	;Used a lot
	CLD
CDLP1:
	XOR	SI,SI
	MOV	DI,SI
	CMP	DX,BX		;More than 64K left?
	JB	CDLP2
	MOV	CX,32768	;32K words
	REP	MOVSW
;Shift everyone by 64K
	MOV	AX,DS
	ADD	AX,BX
	MOV	DS,AX
	MOV	AX,ES
	ADD	AX,BX
	MOV	ES,AX
	SUB	DX,BX
	JMP	CDLP1
CDLP2:	SHL	DX,1		;Calc. words remaining
	SHL	DX,1
	SHL	DX,1
	MOV	CX,DX
	REP	MOVSW
	POP	DI
	POP	SI
	POP	CX
	POP	BX
	POP	AX
	RET
END
